import type b from "@babel/core";
import { type NodePath, type types as t } from "@babel/core";
import invariant from "tiny-invariant";

import type { SupportedTag } from "../macro";
import { macroNames } from "../shared";

import { evaluateArgs, trimPrefix } from "./utils";

import type { BabelPluginUtils } from "./index";

type BabelTypes = typeof b.types;

export function handleMacro({
  t,
  path,
  _,
}: {
  t: BabelTypes;
  path: NodePath<t.Program>;
  _: BabelPluginUtils;
}) {
  const macros = getMacros(t, path, macroNames).map(macro => mapMacro(t, macro));

  for (const { callee, imported, prefix } of macros) {
    const type = imported === "tw" ? "css" : imported === "tws" ? "js" : undefined;
    if (!type) continue;

    if (isNodePath(callee, t.isTaggedTemplateExpression)) {
      const { node } = callee;
      const { quasi } = node;

      invariant(!quasi.expressions.length, `Macro call should not contain expressions`);

      const value = quasi.quasis[0].value.cooked;
      if (value) {
        if (prefix && supportedTags.has(prefix as SupportedTag) && type === "css") {
          const list = trimPrefix(value);
          const className = _.getClass(type, list.join(" "));
          _.recordIfAbsent(type, {
            key: className,
            classNames: list,
            location: _.sliceText(node),
          });
          callee.replaceWith(
            t.callExpression(_.getClassedImport(), [
              t.stringLiteral(prefix),
              t.stringLiteral(className),
            ])
          );
        } else {
          const list = trimPrefix(value, prefix ? prefix + ":" : undefined);
          const className = _.getClass(type, list.join(" "));
          _.recordIfAbsent(type, {
            key: className,
            classNames: list,
            location: _.sliceText(node),
          });
          _.replaceWithImport({
            type,
            path: callee,
            className: addIf(className, list.includes("group") && " group"),
          });
        }
      }
    } else if (isNodePath(callee, t.isCallExpression)) {
      const { node } = callee;
      if (!t.isIdentifier(node.callee)) continue;

      const list = callee.get("arguments").flatMap(evaluateArgs);
      const className = _.getClass("css", list.join(" "));
      _.recordIfAbsent(type, {
        key: className,
        classNames: list,
        location: _.sliceText(node),
      });
      _.replaceWithImport({
        type,
        path: callee,
        className: addIf(className, list.includes("group") && " group"),
      });
    }
  }
}

function addIf(text: string, suffix: string | false) {
  return suffix ? text + suffix : text;
}

function mapMacro(t: BabelTypes, macro: ReturnType<typeof getMacros>[number]) {
  let callee = macro.local.parentPath;
  const prefix: string[] = [];

  while (isNodePath(callee, t.isMemberExpression)) {
    invariant(t.isIdentifier(callee.node.property), "Invalid member expression");
    prefix.unshift(
      callee.node.property.name.replace(/([a-z])([A-Z])/g, "$1-$2").toLowerCase()
    );
    callee = callee.parentPath;
  }

  return {
    callee,
    imported: macro.imported,
    prefix: prefix.length ? prefix.join(":") : undefined,
  };
}

function getName(t: BabelTypes, exp: t.Node) {
  if (t.isIdentifier(exp)) {
    return exp.name;
  } else if (t.isStringLiteral(exp)) {
    return exp.value;
  }
}

function getMacros(
  t: BabelTypes,
  programPath: NodePath<t.Program>,
  importSources: string[]
) {
  const importDecs = programPath
    .get("body")
    .filter(x => isNodePath(x, t.isImportDeclaration))
    .filter(x => importSources.includes(x.node.source.value));

  const macros = importDecs
    .flatMap(x => x.get("specifiers"))
    .map(x => {
      const local = x.get("local");
      if (isNodePath(x, t.isImportNamespaceSpecifier)) {
        return local.scope
          .getOwnBinding(local.node.name)!
          .referencePaths.map(p => p.parentPath)
          .filter(p => isNodePath(p, t.isMemberExpression))
          .map(p => ({
            local: p,
            imported: getName(t, p.node.property)!,
          }))
          .filter(p => p.imported);
      } else if (t.isImportSpecifier(x.node)) {
        const imported = x.node.imported;
        return local.scope.getOwnBinding(local.node.name)!.referencePaths.map(p => ({
          local: p as NodePath<t.Identifier>,
          imported: getName(t, imported)!,
        }));
      }
    })
    .filter(Boolean)
    .flat(1);

  for (const x of importDecs) {
    x.remove();
  }

  return macros;
}

const isNodePath = <T extends t.Node>(
  nodePath: NodePath<t.Node | null | undefined> | null,
  predicate: (node: t.Node) => node is T
): nodePath is NodePath<T> => Boolean(nodePath?.node && predicate(nodePath.node));

const supportedTags = new Set<SupportedTag>([
  "a",
  "abbr",
  "address",
  "article",
  "b",
  "bdi",
  "blockquote",
  "button",
  "caption",
  "cite",
  "code",
  "col",
  "colgroup",
  "dd",
  "del",
  "details",
  "dialog",
  "div",
  "dl",
  "dt",
  "em",
  "fieldset",
  "figcaption",
  "figure",
  "footer",
  "form",
  "h1",
  "h2",
  "h3",
  "h4",
  "h5",
  "h6",
  "header",
  "hr",
  "i",
  "iframe",
  "img",
  "input",
  "ins",
  "kbd",
  "label",
  "legend",
  "li",
  "main",
  "mark",
  "menu",
  "meter",
  "nav",
  "ol",
  "output",
  "p",
  "picture",
  "pre",
  "progress",
  "q",
  "rt",
  "ruby",
  "section",
  "select",
  "span",
  "strong",
  "sub",
  "summary",
  "sup",
  "svg",
  "table",
  "tbody",
  "td",
  "textarea",
  "tfoot",
  "th",
  "thead",
  "time",
  "tr",
  "u",
  "ul",
  "var",
  "video",
  "wbr",
]);
