// https://github.com/adobe/react-spectrum/blob/14f324fe890fcedc6e34889d9b04d5d6bfeb8380/packages/tailwindcss-react-aria-components/src/index.js
import plugin from "tailwindcss/plugin.js";
import type { PluginAPI } from "tailwindcss/types/config";

// Order of these is important because it determines which states win in a conflict.
// We mostly follow Tailwind's defaults, adding our additional states following the categories they define.
// https://github.com/tailwindlabs/tailwindcss/blob/304c2bad6cb5fcb62754a4580b1c8f4c16b946ea/src/corePlugins.js#L83
const attributes = {
  boolean: [
    // Conditions
    "allows-removing",
    "allows-sorting",
    "allows-dragging",
    "has-submenu",

    // States
    "open",
    "entering",
    "exiting",
    "indeterminate",
    ["placeholder-shown", "placeholder"],
    "current",
    "required",
    "unavailable",
    "invalid",
    ["read-only", "readonly"],
    "outside-month",
    "outside-visible-range",

    // Content
    "empty",

    // Interactive states
    "focus-within",
    ["hover", "hovered"],
    ["focus", "focused"],
    "focus-visible",
    "pressed",
    "selected",
    "selection-start",
    "selection-end",
    "dragging",
    "drop-target",
    "resizing",
    "disabled",
  ] as const,
  enum: {
    placement: ["left", "right", "top", "bottom"],
    type: ["literal", "year", "month", "day"],
    layout: ["grid", "stack"],
    orientation: ["horizontal", "vertical"],
    "selection-mode": ["single", "multiple"],
    "resizable-direction": ["right", "left", "both"],
    "sort-direction": ["ascending", "descending"],
  },
};

const shortNames: Record<string, string> = {
  "selection-mode": "selection",
  "resizable-direction": "resizable",
  "sort-direction": "sort",
};

// Variants we use that are already defined by Tailwind:
// https://github.com/tailwindlabs/tailwindcss/blob/a2fa6932767ab328515f743d6188c2164ad2a5de/src/corePlugins.js#L84
const nativeVariants = [
  "indeterminate",
  "required",
  "invalid",
  "empty",
  "focus-visible",
  "focus-within",
  "disabled",
];
const nativeVariantSelectors = new Map<string, string>([
  ...nativeVariants.map(variant => [variant, `:${variant}`] as const),
  ["hovered", ":hover"],
  ["focused", ":focus"],
  ["readonly", ":read-only"],
  ["open", "[open]"],
]);

// Variants where both native and RAC attributes should apply. We don't override these.
const nativeMergeSelectors = new Map([["placeholder", ":placeholder-shown"]]);

type SelectorFn = (wrap: (s: string) => string) => string;
type SelectorValue = string | SelectorFn;
type Selector = string | [string, SelectorValue];

// If no prefix is specified, we want to avoid overriding native variants on non-RAC components, so we only target elements with the data-rac attribute for those variants.
function getSelector(
  prefix: string,
  attributeName: string,
  attributeValue: string | null,
  hoverOnlyWhenSupported: boolean
): Selector {
  const baseSelector = attributeValue
    ? `[data-${attributeName}="${attributeValue}"]`
    : `[data-${attributeName}]`;
  const nativeSelector = nativeVariantSelectors.get(attributeName);
  if (prefix === "" && nativeSelector) {
    const wrappedNativeSelector = `&:not([data-rac])${nativeSelector}`;
    let nativeSelectorGenerator: SelectorValue = wrappedNativeSelector;
    if (nativeSelector === ":hover" && hoverOnlyWhenSupported) {
      nativeSelectorGenerator = wrap =>
        `@media (hover: hover) and (pointer: fine) { ${wrap(wrappedNativeSelector)} }`;
    }
    return [`&[data-rac]${baseSelector}`, nativeSelectorGenerator];
  } else if (prefix === "" && nativeMergeSelectors.has(attributeName)) {
    return [`&${baseSelector}`, `&${nativeMergeSelectors.get(attributeName)}`];
  } else {
    return `&${baseSelector}`;
  }
}

const mapSelector = (selector: Selector, fn: (v: SelectorValue) => string) =>
  Array.isArray(selector) ? selector.map(fn) : fn(selector);

const wrapSelector = (selector: SelectorValue, wrap: (text: string) => string) =>
  typeof selector === "function" ? selector(wrap) : wrap(selector);

const addVariants = (
  variantName: string,
  selectors: Selector,
  addVariant: PluginAPI["addVariant"],
  matchVariant: PluginAPI["matchVariant"]
) => {
  addVariant(
    variantName,
    mapSelector(selectors, selector => wrapSelector(selector, s => s))
  );
  matchVariant(
    "group",
    (_, { modifier }) =>
      modifier
        ? mapSelector(selectors, selector =>
            wrapSelector(selector, s => `:merge(.group\\/${modifier})${s.slice(1)} &`)
          )
        : mapSelector(selectors, selector =>
            wrapSelector(selector, s => `:merge(.group)${s.slice(1)} &`)
          ),
    { values: { [variantName]: variantName } }
  );
  matchVariant(
    "peer",
    (_, { modifier }) =>
      modifier
        ? mapSelector(selectors, selector =>
            wrapSelector(selector, s => `:merge(.peer\\/${modifier})${s.slice(1)} ~ &`)
          )
        : mapSelector(selectors, selector =>
            wrapSelector(selector, s => `:merge(.peer)${s.slice(1)} ~ &`)
          ),
    { values: { [variantName]: variantName } }
  );
};

export default plugin.withOptions<{ prefix: string }>(
  options =>
    ({ addVariant, matchVariant, config }) => {
      const prefix = options?.prefix ? `${options.prefix}-` : "";
      const future = config().future;
      const hoverOnlyWhenSupported: boolean =
        future === "all" ||
        (typeof future === "object" &&
          !Array.isArray(future) &&
          !!future?.hoverOnlyWhenSupported);

      // Enum attributes go first because currently they are all non-interactive states.
      for (const [attributeName, value] of Object.entries(
        attributes.enum
      ) as (keyof typeof attributes.enum)[][]) {
        for (const attributeValue of value) {
          const name = shortNames[attributeName] || attributeName;
          const variantName = `${prefix}${name}-${attributeValue}`;
          const selectors = getSelector(
            prefix,
            attributeName,
            attributeValue,
            hoverOnlyWhenSupported
          );
          addVariants(variantName, selectors, addVariant, matchVariant);
        }
      }

      for (const attribute of attributes.boolean) {
        let variantName = Array.isArray(attribute) ? attribute[0] : attribute;
        variantName = `${prefix}${variantName}`;
        const attributeName = Array.isArray(attribute) ? attribute[1] : attribute;
        const selectors = getSelector(
          prefix,
          attributeName,
          null,
          hoverOnlyWhenSupported
        );
        addVariants(variantName, selectors, addVariant, matchVariant);
      }
    }
);
